import atexit
import heapq
import threading
import time
import traceback
from concurrent.futures import ThreadPoolExecutor

_g_lock = threading.RLock()
_schedulers = []


def _close_schedulers():
    with _g_lock:
        while _schedulers:
            try:
                s = _schedulers.pop()
                s.close()
            except:
                traceback.print_exc()


atexit.register(_close_schedulers)


class ThreadPoolScheduler:
    def __init__(self, max_threads=5):
        self._task_queue = []
        self._lock = threading.Lock()
        self._scheduler_event = threading.Event()
        self._counter = 0
        self._running = True
        self._scheduler_thread = threading.Thread(target=self._scheduler_loop, name="thread-pool-sch", daemon=True)
        self._executor = ThreadPoolExecutor(max_workers=max_threads, thread_name_prefix="thread-pool-exec")
        with _g_lock:
            _schedulers.append(self)
        self._scheduler_thread.start()

    def _scheduler_loop(self):
        while self._running:
            if self._task_queue:
                execute_at, _, task = self._task_queue[0]
                current_time = time.time()

                if execute_at > current_time:
                    # Head is still in future. Sleep
                    timeout = execute_at - current_time
                    self._scheduler_event.wait(timeout)
                    self._scheduler_event.clear()
                    continue

                # Ready to execute the head
                heapq.heappop(self._task_queue)
                self._executor.submit(task)
            else:
                # queue is empty.
                self._scheduler_event.wait()
                self._scheduler_event.clear()

    def submit(self, target_function, delay_seconds):
        with self._lock:
            execute_at = time.time() + delay_seconds
            # tuple subitems are compared after comparing the first one.
            # So make sure there's a differentiator with counter.
            self._counter += 1
            heapq.heappush(self._task_queue, (execute_at, self._counter, target_function))
            self._scheduler_event.set()

    def close(self):
        self._running = False
        self._scheduler_event.set()
        self._executor.shutdown(wait=False, cancel_futures=True)
        with _g_lock:
            try:
                _schedulers.remove(self)
            except:
                pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


# Example usage
# def print_message(message):
#     print(message)
#     time.sleep(5)
#     print("done")

# scheduler = ThreadPoolScheduler(max_threads=5)
# _n = 0

# def rr(txt):
#     global _n
#     print("rr", txt, _n)
#     _n += 1
#     scheduler.submit(lambda: rr("a1"), 0)
# scheduler.submit(lambda: rr("a1"), 0)
# time.sleep(5)  # Give some time for tasks to execute



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re
from datetime import datetime, timedelta, timezone

PATTERN = re.compile("^(([0-9]+)D)?(([0-9]+)H)?(([0-9]+)M)?(([0-9]+)S)?$")


# Examples

# 11D12H13M14S
# 2H33S


def to_seconds(duration_string: str) -> int:
    if not duration_string:
        return 0
    duration_string = duration_string.upper()

    future = duration_string[0] == "-"
    if future:
        duration_string = duration_string[1:]

    matcher = PATTERN.fullmatch(duration_string)
    if matcher is None:
        raise Exception("Unsupported duration format: %s" % duration_string)

    d = matcher.group(2)
    h = matcher.group(4)
    m = matcher.group(6)
    s = matcher.group(8)

    total_seconds = 0
    if d:
        total_seconds += int(d) * 24 * 3600
    if h:
        total_seconds += int(h) * 3600
    if m:
        total_seconds += int(m) * 60
    if s:
        total_seconds += int(s)

    if future:
        total_seconds = -total_seconds
    return total_seconds


def to_duration(seconds: int, simple: bool = False) -> str:

    future = seconds < 0
    if future:
        seconds = -seconds

    ONE_MINUTE = 60
    ONE_HOUR = ONE_MINUTE * 60
    ONE_DAY = ONE_HOUR * 24
    days = int(seconds / ONE_DAY)
    seconds %= ONE_DAY
    hours = int(seconds / ONE_HOUR)
    seconds %= ONE_HOUR
    minutes = int(seconds / ONE_MINUTE)
    seconds %= ONE_MINUTE

    ret = ""
    v = 0
    while True:
        if days > 0:
            ret += f"{days}D"
            v += 1
        if hours > 0:
            ret += f"{hours}H"
            v += 1
            if simple and v >= 2:
                break
        if minutes > 0:
            ret += f"{minutes}M"
            v += 1
            if simple and v >= 2:
                break
        if seconds > 0:
            ret += f"{seconds}S"
        break
    if future:
        ret = "-" + ret
    return ret


def format_timedelta(d: timedelta, simple: bool = True):
    seconds = d.days * 86400 + d.seconds  # + d.microseconds / 1e6
    return to_duration(seconds, simple)


def to_utc(dt):
    if isinstance(dt, int) or isinstance(dt, float):
        dt = datetime.fromtimestamp(dt)
    if isinstance(dt, str):
        dt = datetime.fromisoformat(dt)

    if dt.tzinfo is None:
        # Naive datetime: Assume it is in the local timezone and convert to UTC
        dt = dt.replace(tzinfo=timezone.utc)  # Assuming naive datetime is in UTC for this example
    return dt.astimezone(timezone.utc)


def stale(when) -> str:
    return format_timedelta(from_now(when), True)


def from_now(when) -> timedelta:
    when_utc = to_utc(when)
    now_utc = datetime.now().astimezone(timezone.utc)
    return now_utc - when_utc


def _test():
    data = {
        "14S": 14,
        "13M": 13 * 60,
        "12H": 12 * 3600,
        "11D": 11 * 24 * 3600,
        "13M14S": 13 * 60 + 14,
        "12H14S": 12 * 3600 + 14,
        "11D14S": 11 * 24 * 3600 + 14,
        "12H13M": 12 * 3600 + 13 * 60,
        "11D13M": 11 * 24 * 3600 + 13 * 60,
        "11D12H": 11 * 24 * 3600 + 12 * 3600,
        "11D12H13M": 11 * 24 * 3600 + 12 * 3600 + 13 * 60,
        "11D12H14S": 11 * 24 * 3600 + 12 * 3600 + 14,
        "11D13M14S": 11 * 24 * 3600 + 13 * 60 + 14,
        "12H13M14S": 12 * 3600 + 13 * 60 + 14,
        "11D12H13M14S": 11 * 24 * 3600 + 12 * 3600 + 13 * 60 + 14,
    }
    for k in data.keys():
        v = data[k]
        assert to_seconds(k) == v, f"to_seconds failed: k={k},v={v},got={to_seconds(k)}"
        assert to_duration(v) == k, f"to_duration failed: k={k},v={v},got={to_duration(v)}"
    print("PASS")


if __name__ == "__main__":
    _test()



from datetime import datetime, timezone


def iso_date_to_timestamp(datetime_string: str) -> int:
    dt_object = datetime.strptime(datetime_string, "%Y-%m-%dT%H:%M:%S.%fZ")
    return int(dt_object.replace(tzinfo=timezone.utc).timestamp() * 1000)


def timestamp_to_iso_date(timestamp_ms: int) -> str:
    dt = datetime.fromtimestamp(timestamp_ms / 1000, tz=timezone.utc)
    return dt.isoformat(timespec="milliseconds").replace("+00:00", "Z")



import atexit
import logging

log = logging.getLogger(__name__)
_sg_client = None


# Work around issue:
# http11.py, line 211, in close: AttributeError: 'NoneType' object has no attribute 'CLOSED'
def _on_exit():
    if _sg_client:
        _sg_client.close()

    log.info("Exit")


atexit.register(_on_exit)



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import os
import shutil
from collections.abc import Generator
from os import listdir, path
from typing import Any

from . import jsondot
from .jsondot import dotdict

log = logging.getLogger(__name__)


def _validate_key(key: str):
    # if key.find(os.pathsep) >= 0:
    if key.find("/") >= 0 or key.find("\\") >= 0:
        raise Exception("Invalid key: " + key)


def _load_yaml(file_name: str):
    if not os.path.exists(file_name):
        return

    import yaml

    with open(file_name, "r") as file:
        ret = yaml.safe_load(file)
        return jsondot.dotify(ret)


def _load_text(file_name: str) -> str:
    if not os.path.exists(file_name):
        return
    with open(file_name, "r", encoding="utf-8") as file:
        return file.read()


def _load_file(file_path: str, format: str):
    try:
        _, ext = path.splitext(file_path)

        if ext == ".json" or (ext == "" and format == "json"):
            return jsondot.load(file_path, lock=True)
        if ext == ".yaml" or ext == ".yml" or (ext == "" and (format == "yaml" or format == "yml")):
            return _load_yaml(file_path)
        if ext == ".txt" or (ext == "" and format in ["text", "plain", "txt"]):
            return _load_text(file_path)

        if format != "auto":
            raise Exception(f"Invalid store format: {format}")

        # The specified path has no ext name, and format is auto.
        if os.path.exists(file_path):
            return jsondot.load(file_path, lock=True)

        tmp = file_path + ".json"
        if os.path.exists(tmp):
            return jsondot.load(file_path + ".json", lock=True)

        tmp = file_path + ".yaml"
        if os.path.exists(tmp):
            return _load_yaml(tmp)

        tmp = file_path + ".yml"
        if os.path.exists(tmp):
            return _load_yaml(tmp)

        # Not found
    except Exception as e:
        raise Exception(f"Fail with file {file_path}") from e


def _list_sub_dirs(directory, depth):
    def recursive_list(current_dir, current_depth, base_dir):
        if depth != -1 and current_depth > depth:
            return []
        subdirs = []
        with os.scandir(current_dir) as entries:
            for entry in entries:
                if entry.is_dir():
                    rel_path = os.path.relpath(entry.path, base_dir)
                    subdirs.append(rel_path)
                    subdirs.extend(recursive_list(entry.path, current_depth + 1, base_dir))
        return subdirs

    return recursive_list(directory, 0, directory)


class _ManagedDict(dict):
    def __init__(self, initial_data: dict, store: "fstore", name: str):
        super().__init__(initial_data)
        self._store = store
        self._name = name

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._store.save(self._name, self)


class fstore:
    """A key-value store, optionally backed by a file system directory and files in it."""

    def __init__(self, store_path: str = None, create: bool = True):
        """Initialize the store

        Args:
                store_path (str): The path to store state files. If None, state will not be stored.
                create (bool, optional): If store_path is specified, try creating it if not exist. Defaults to True.
        """
        if store_path:
            self._path = path.realpath(store_path)
            self._create = create
            self._path_checked = False
        else:
            self._path = None
        self._cache = {}

    def get(self, key: str, reload: bool = False, format: str = "auto", default=None) -> dotdict:
        _validate_key(key)
        if self._path and (reload or key not in self._cache):
            file_path = self._get_path(key)
            log.debug(f"Read {file_path}")
            data = _load_file(file_path, format)
            if data is not None:
                self._cache[key] = data
        else:
            data = self._cache.get(key)

        if data is None and default is not None:
            return jsondot.dotify(default)
        return data

    def exists(self, key: str) -> bool:
        _validate_key(key)
        if key in self._cache:
            return True
        return path.exists(path.join(self._path, key))

    def _get_path(self, key: str) -> str:
        if not self._path:
            return None
        return path.join(self._path, key)

    def _ensure_dir(self):
        if not self._path or self._path_checked:
            return
        if path.exists(self._path):
            if not os.path.isdir(self._path):
                raise Exception(f"Store path is not a directory: {self._path}")
            else:
                # Good.
                self._path_checked = True
                return
        else:
            if self._create:
                os.makedirs(self._path, exist_ok=True)
            else:
                raise Exception(f"Store path does not exist: {self._path}")
        self._path_checked = True

    def save(self, key: str, data: Any, format: str = "auto") -> Any:
        _validate_key(key)

        if format == "auto":
            if isinstance(data, str):
                format = "text"
            else:
                format = "json"

        data = jsondot.dotify(data)

        self._cache[key] = data
        if self._path:
            self._ensure_dir()
            file_path = self._get_path(key)
            log.debug(f"Write {file_path}")
            if format == "json":
                jsondot.save(data, file_path, lock=True)
            elif format == "text":
                with open(file_path, "w", encoding="utf-8") as outfile:
                    # TODO lock
                    outfile.write(str(data))
            elif format == "json-compact":
                jsondot.save(data=data, file=file_path, pretty=False, lock=True)
            elif format == "yaml":
                import yaml

                with open(file_path, "w", encoding="utf-8") as outfile:
                    # TODO lock
                    yaml.safe_dump(data, outfile)
            else:
                raise Exception(f"Invalid format. key={key}, format={format}")

        return data

    def delete(self, key: str) -> None:
        key = str(key)
        _validate_key(key)
        self._cache.pop(key, None)
        if self._path:
            file_path = self._get_path(key)
            if os.path.exists(file_path):
                os.remove(file_path)

    def keys(self) -> list[str]:
        if self._path:
            if not os.path.exists(self._path):
                return []
            return [f for f in listdir(self._path) if path.isfile(path.join(self._path, f))]
        return list(self._cache.keys())

    def children(self, depth: int = 0) -> list[str]:
        """List child stores"""
        if self._path:
            if not os.path.exists(self._path):
                return []
            if depth == 0:
                return [f for f in listdir(self._path) if path.isdir(path.join(self._path, f))]
            else:

                return _list_sub_dirs(self._path, depth)

        raise Exception("TODO: list children for RAM store")

    def child(self, name) -> "fstore":
        return fstore(self._path + "/" + name)

    def values(self) -> Generator[Any, None, None]:
        for k in self.keys():
            yield self.get(k)

    def items(self) -> Generator[tuple[str, dict], None, None]:
        for k in self.keys():
            yield (k, self.get(k))

    def clear(self) -> None:
        for k in self.keys():
            self.delete(k)

    def destroy(self) -> None:
        self._cache.clear()
        if self._path:
            shutil.rmtree(self._path)

    def size(self) -> int:
        return len(self.keys())

    def contains(self, key: str) -> bool:
        if key in self._cache:
            return True
        return key in self.keys()

    # ----------------- Helpers -----------------

    def patch(self, key: str, data: dict) -> dotdict:
        existing_data = self.get(key)
        if existing_data is None:
            existing_data = jsondot.dotify({})
        existing_data |= data
        self.save(key, existing_data)
        return existing_data

    def doc(self, key: str) -> dict:
        data = self.get(key, {})
        return _ManagedDict(data, self, key)



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import logging
import sys
from http.client import HTTPResponse
from typing import Callable, Optional, Type, Union

import httpx
from authlib.integrations.httpx_client import OAuth2Client
from pydantic import BaseModel

from hcs_core.ctxp import jsondot

log = logging.getLogger(__name__)

_print_http_error = True


def parse(resp: httpx.Response):
    if not resp.content:
        return None
    content_type = resp.headers["Content-Type"]
    if content_type.startswith("text"):
        return resp.text
    if len(resp.content) > 3:
        try:
            data = resp.json()
            if isinstance(data, list):
                return [jsondot.dotify(obj) for obj in data]
            elif isinstance(data, dict):
                return jsondot.dotify(data)
        except:
            log.info("--- Fail parsing json. Dump content ---")
            log.info(resp.content)
            raise
    else:
        return None


def _raise_on_4xx_5xx(response: httpx.Response):
    if not response.is_success:
        response.read()
        if len(response.text) > 0:
            text = _try_formatting_json(response.text)
            log.debug(text)

    response.raise_for_status()


def _try_formatting_json(text: str):
    try:
        return json.dumps(json.loads(text), indent=4)
    except:
        return text


def _log_request(request):
    if not log.isEnabledFor(logging.DEBUG):
        return
    log.debug("\n")
    log.debug(f"--> {request.method} {request.url}")
    log.debug(f"--> {request.headers}")

    if len(request.content) > 0:
        text = _try_formatting_json(request.content)
        log.debug(text)


def _log_response(response: httpx.Response):
    if not log.isEnabledFor(logging.DEBUG):
        return
    request = response.request
    log.debug(f"<-- {request.method} {request.url} - {response.status_code}")
    log.debug(f"<-- {request.headers}")
    response.read()
    if len(response.text) > 0:
        text = _try_formatting_json(response.text)
        log.debug(text)
    log.debug("\n")


def _parse_resp(resp: httpx.Response):
    if not resp.content:
        return
    content_type = resp.headers["Content-Type"]
    if content_type.startswith("text"):
        return resp.text
    if content_type == "application/json" and resp.content:
        try:
            data = resp.json()
            return jsondot.dotify(data)
        except:
            log.info("--- Fail parsing json. Dump content ---")
            log.info(resp.content)
            raise
    return resp.text


def _is_404(e: httpx.HTTPStatusError) -> bool:
    return e.response.status_code == 404


def on404ReturnNone(func):
    try:
        resp = func()
        return _parse_resp(resp)
    except httpx.HTTPStatusError as e:
        if e.response.status_code == 404:
            return None
        raise


def _raise_http_error(e: httpx.HTTPStatusError):
    if _print_http_error:
        print(e.response.text, file=sys.stderr)
    raise e


class EzClient:
    def __init__(
        self, base_url: Union[str, Callable], oauth_client: OAuth2Client = None, lazy_oauth_client: Callable = None
    ) -> None:
        # self._client = httpx.Client(base_url=base_url, timeout=30, event_hooks=event_hooks)
        self._base_url = base_url
        self._client_impl = oauth_client
        self._lazy_oauth_client = lazy_oauth_client

    def _client(self):
        if not self._client_impl:
            client = self._lazy_oauth_client()
            base_url = self._base_url() if callable(self._base_url) else self._base_url
            client.base_url = base_url
            client.timeout = 30
            request_hooks = client.event_hooks["request"]
            response_hooks = client.event_hooks["response"]
            if _log_request not in request_hooks:
                request_hooks.append(_log_request)
            if _log_response not in response_hooks:
                response_hooks.append(_log_response)
            if _raise_on_4xx_5xx not in response_hooks:
                response_hooks.append(_raise_on_4xx_5xx)
            self._client_impl = client
        return self._client_impl

    def post(
        self,
        url: str,
        json: dict = None,
        text: str = None,
        files=None,
        headers: dict = None,
        type: Optional[Type[BaseModel]] = None,
    ):
        # import json as jsonlib
        # print(url, jsonlib.dumps(json, indent=4))
        try:
            resp = self._client().post(url, json=json, content=text, files=files, headers=headers)
        except httpx.HTTPStatusError as e:
            _raise_http_error(e)
        data = _parse_resp(resp)
        if data and type:
            try:
                return type.model_validate(data)
            except:
                log.info("--- Fail converting model. Dump content ---")
                log.info(data)
                raise
        return data

    def get(self, url: str, raise_on_404: bool = False, type: Optional[Type[BaseModel]] = None):
        try:
            # print("->", url)
            resp = self._client().get(url)
            data = _parse_resp(resp)
            if data and type:
                try:
                    return type.model_validate(data)
                except:
                    log.info("--- Fail converting model. Dump content ---")
                    log.info(data)
                    raise
            return data
        except httpx.HTTPStatusError as e:
            if _is_404(e):
                if raise_on_404:
                    _raise_http_error(e)
                else:
                    pass
            else:
                _raise_http_error(e)

    def patch(self, url: str, json: dict):
        try:
            resp = self._client().patch(url, json=json)
            return _parse_resp(resp)
        except httpx.HTTPStatusError as e:
            _raise_http_error(e)
            raise

    def delete(self, url: str, raise_on_404: bool = False):
        try:
            resp = self._client().delete(url)
            return _parse_resp(resp)
        except httpx.HTTPStatusError as e:
            if _is_404(e):
                if raise_on_404:
                    _raise_http_error(e)
                else:
                    pass
            else:
                _raise_http_error(e)

    def put(self, url: str, json: dict):
        try:
            resp = self._client().put(url, json=json)
            return _parse_resp(resp)
        except httpx.HTTPStatusError as e:
            _raise_http_error(e)

    def close(self):
        self._client().close()

    def dump_response(self, response: HTTPResponse):
        log.info("response text: " + response.text())



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import re
from typing import Any

from . import jsondot
from .profile_store import fstore, profile_store


def _store() -> fstore:
    return profile_store("context")


def list() -> list:
    return _store().keys()


def get(name: str, reload: bool = False, default=None) -> jsondot.dotdict:
    return _store().get(key=name, reload=reload, default=default)


def set(name: str, data: dict):
    return _store().save(name, data)


def delete(name: str):
    return _store().delete(name)


def file(name: str):
    return _store()._get_path(name)


def clear():
    return _store().clear()


class Context:
    def __init__(self, name):
        self.name = name
        self._store = _store()
        self._data = None
        self._changed = False

    def get(self, key: str, default=None):
        return self._data.get(key, default)

    def set(self, key: str, value: Any):
        existing = self._data.get(key)
        if existing != value:
            self._changed = True
            self._data[key] = value
        return self

    def remove(self, key: str):
        if key in self._data:
            self._changed = True
            del self._data[key]

    def __enter__(self):
        self._data = self._store.get(key=self.name, reload=False, default={})

    def __exit__(self, exc_type, exc_value, traceback):
        if self._changed:
            self._store.save(key=self.name, data=self.data)
        return False

    def apply_variables(self, object: any, additional_vars: dict = None):

        mapping = self._data.copy()
        if additional_vars:
            mapping.update(additional_vars)

        # The Python default Template.substitute has too many limitations...
        def substitute(text: str):
            pattern = re.compile(r".*\${(.+?)}.*")
            while True:
                m = pattern.match(text)
                if not m:
                    break
                name = m.group(1)
                text = text.replace("${" + name + "}", self.get(name))
            return text

        t = type(object)
        try:
            if t is str:
                return substitute(object)

            if t is dict:
                text = json.dumps(object)
                text = substitute(text)
                return json.loads(text)

            if t is jsondot.dotdict:
                text = json.dumps(object)
                text = substitute(text)
                return jsondot.parse(text)
        except KeyError as e:
            raise Exception("Variable not defined in context") from e

    def load_template(self, name: str):
        # with open(name) as f:
        #     text = f.read()
        # return apply_variables(text)

        tmpl = jsondot.load(name)
        return self.apply_variables(tmpl)
